#include <stdio.h>
#include <string.h>
#include <assert.h>

#include "../include/oberon.h"

static char source_test[] =
	"(* Main module *)"
	"MODULE Test;"
	"IMPORT Out;"
	""
	"VAR"
	"  i, len : INTEGER;"
	""
	"BEGIN"
	"  Out.Open();"
	"  len := 2 * 8;"
	"  FOR i := 0 TO len BY 2 DO"
	"    Out.String('Count '); Out.Int(i, 0); Out.Ln;"
	"    len := len + 2;"
	"  END;"
	"  Out.String('end'); Out.Ln;"
	"END Test."
;

static char source_out[] =
	"MODULE Out;"
	"  PROCEDURE Open*;"
	"  END Open;"
	""
	"  PROCEDURE Char* (ch : CHAR);"
	"  END Char;"
	""
	"  PROCEDURE String* (str : ARRAY OF CHAR);"
	"  END String;"
	""
	"  PROCEDURE Int*(i, n : LONGINT);"
	"  END Int;"
	""
	"  PROCEDURE Real*(x : REAL; n : INTEGER);"
	"  END Real;"
	""
	"  PROCEDURE LongReal*(x : LONGREAL; n : INTEGER);"
	"  END LongReal;"
	""
	"  PROCEDURE Ln*;"
	"  END Ln;"
	""
	"END Out."
;

static oberon_context_t * ctx;
static oberon_module_t * mod;

static const char *
import_module(const char * name)
{
	if(strcmp(name, "Test") == 0)
	{
		return source_test;
	}
	else if(strcmp(name, "Out") == 0)
	{
		return source_out;
	}
	else
	{
		return NULL;
	}
}

typedef void (*TOutOpen)();
static TOutOpen * OutOpenPtr;
void ImplOutOpen()
{
}

typedef void (*TOutInt)(int, int);
static TOutInt * OutIntPtr;
void ImplOutInt(int i, int n)
{
	char number[22];
	snprintf(number, 22, "%d", i);
	int len = strlen(number);
	for(int i = 0; i < n - len; i++)
	{
		putchar(' ');
	}
	printf("%s", number);
}

typedef void (*TOutReal)(float, int);
static TOutReal * OutRealPtr;
void ImplOutReal(float i, int n)
{
	char number[32];
	snprintf(number, 32, "%F", i);
	int len = strlen(number);
	for(int i = 0; i < n - len; i++)
	{
		putchar(' ');
	}
	printf("%s", number);
}

typedef void (*TOutLn)();
static TOutLn * OutLnPtr;
void ImplOutLn()
{
	putchar('\n');
}

void init_system_modules()
{
	OutOpenPtr = oberon_generator_get_var(ctx, "Out_Open");
	*OutOpenPtr = ImplOutOpen;
	OutIntPtr = oberon_generator_get_var(ctx, "Out_Int");
	*OutIntPtr = ImplOutInt;
	OutRealPtr = oberon_generator_get_var(ctx, "Out_Real");
	*OutRealPtr = ImplOutReal;
	OutLnPtr = oberon_generator_get_var(ctx, "Out_Ln");
	*OutLnPtr = ImplOutLn;
}

void start_module()
{
	void (*begin)() = oberon_generator_get_procedure(ctx, "Test_BEGIN");
	begin();
}

int
main(int argc, char ** argv)
{
	ctx = oberon_create_context(import_module);
	mod = oberon_compile_module(ctx, source_test);

	oberon_generate_code(ctx);

//	init_system_modules();

//	oberon_generator_dump(ctx, "dump.txt");

//	start_module();

	oberon_destroy_context(ctx);
	return 0;
}
